/**
* @author patrick
*/

var fs = require('fs');

var nnHelper = function(){
	this.writeDebug = function(s){
		if(this.debug==true){
			console.log(s);
		}
	}

	this.handleError = function(s){
		console.log(s);
	}
}

var helper = new nnHelper();


var nnNetwork = function(){
	this.nodes = Array();
	this.debug = false;

	//just pass an array that defines the layers... element 0 is input
	//example is Array(3,2,1) for a nn with input 3, hidden 2, and output 1
	this.initialize = function(dimensions){
		var i,j;
		for(i=0; i < dimensions.length; i++){
			this.nodes[i] = Array();
			for(j=0; j < dimensions[i]; j++){
				this.nodes[i][j]=new nnNode();
				helper.writeDebug("creating node at "+i+", "+j);
			}
		}
	}

	//will load from file that defines the neural network
	this.initializeFromFile=function(filename){
		var i,j;
		var infile = JSON.parse(fs.readFileSync(filename,'utf8'));
		for(i=0; i<infile.nodes.length; i++){
			this.nodes[i]=Array();
			for(j=0; j< infile.nodes[i].length; j++){
				this.nodes[i][j]=new nnNode();
				this.nodes[i][j].setWeights(infile.nodes[i][j].weights);
				this.nodes[i][j].setBias(infile.nodes[i][j].bias);
				this.nodes[i][j].coef = infile.nodes[i][j].coef;
				this.nodes[i][j].aBias = infile.nodes[i][j].aBias;
			}
		}		
		helper.writeDebug(this.nodes[2][0].weights[0]);
	}

	this.writeFile=function(filename){
		var i;
		outj = {};
		outj.structure=Array();
		//add the structure
		for(i=0; i<this.nodes.length; i++){
			outj.structure[i]=this.nodes[i].length;
		}


		//write json to file
		fs.writeFile(filename, JSON.stringify(this, null, 4), function(err){
			if(err) {
				helper.handleError("unable to write to the file:"+err);
			}
		});
	}

	//sets the values of the input layer, takes an array
	this.inputLayer = function(values){
		var i;
		if(values.length!=this.nodes[0].length){
			this.error("attempting to set input layer with invalid number of values");
		}else{
			for(i=0; i<values.length; i++){
				this.nodes[0][i].output = values[i];
				this.nodes[0][i].isInputLayer=true;
				helper.writeDebug("setting input "+i+" to "+values[i]);
			}
		}
	}

	this.getOutputLayer = function(){
		return this.nodes[this.nodes.length-1];
	}

	this.getOutputLayerValues = function(){
		var ret = Array();
		var i, layer;
		layer = this.nodes.length-1;
		for(i=0; i< this.nodes[layer].length; i++){
			ret[i]=this.nodes[layer][i].output;
		}
		return ret;
	}

	//sets random weights
	this.randomWeights = function(){
		var i,j,w;
		for(i=1; i<this.nodes.length; i++){
			for(j=0; j<this.nodes[i].length; j++){
				weightArray=Array();
				for(w=0;w<this.nodes[i-1].length;w++){
					weightArray[w]=Math.floor(Math.random() * 100.0) / 100.0;
				}
				this.nodes[i][j].setWeights(weightArray);
				this.nodes[i][j].setBias((Math.floor(Math.random() * 100.0) / 100.0));
			}
		}
	}


	this.computeOut = function(){
		var i,j,k;
		helper.writeDebug("computing!");

		for(i=1; i<this.nodes.length; i++){
			helper.writeDebug("computing output for layer "+i)
			for(j=0;j<this.nodes[i].length;j++){
				inArray=Array();
				for(k=0;k<this.nodes[i-1].length;k++){
					helper.writeDebug("getting output for "+(i-1)+"|"+k);

					inArray[k]=this.nodes[i-1][k].getOutput();



				}
				this.nodes[i][j].setInputs(inArray);
				this.nodes[i][j].getOutput();
				helper.writeDebug("inputs set for"+ i + ":" + j);
				helper.writeDebug(JSON.stringify(this.nodes[i][j], null, 4));
			}
		}
	}

	//this adjusts the network, t is the desired output array (must match output layer), a good value for n is .1
	this.train = function(t, n){
		var i,j, layer;
		this.computeOut();
		helper.writeDebug("starting training " + t[0]);
		target=this.nodes.length-1;
		if(t.length!=this.nodes[target].length){
			helper.handleError("target "+ t.length+" does not match output layer count "+this.nodes[target].length);
		}

		// //compute output on input layer
		// for(j=0; j<this.nodes[0].length; j++){
			// 	this.nodes[0][j].calcOutput();
			// }

			//start from output layer
			for(i=0; i < this.nodes[target].length; i++){
				currentOut=this.nodes[target][i].getOutput();
				helper.writeDebug( "node # " + t[i] + " : " + this.nodes[target][i].coef + " * " + currentOut + "( * )" + (1-currentOut) + " * " + t[i] + " - " + currentOut);
				this.nodes[target][i].setError( this.nodes[target][i].coef * (currentOut * (1-currentOut) * (t[i]-currentOut)) );
				helper.writeDebug("set "+target+"|"+i+" error to "+this.nodes[target][i].errorVal);
				this.nodes[target][i].calcABias(n);
				this.nodes[target][i].calcAWeights(n);

			}

			//find error and adjust all the rest
			for(layer=(target-1); layer>0; layer--){
				helper.writeDebug("adjusting.. going into layer " + layer);
				for(i=0; i<this.nodes[layer].length; i++){
					helper.writeDebug("!!!!!!" + i);
					//we need to know the sum of errors in the node after this * weights to this node
					//sum(wn * errorn)
					esum=0;
					for(j=0; j<this.nodes[layer+1].length; j++){
						esum+=this.nodes[(layer+1)][j].getSingleError(i);
					}
					helper.writeDebug("sum of errors is "+ esum);

					//compute error of the node
					currentOut=this.nodes[layer][i].getOutput();
					this.nodes[layer][i].setError(this.nodes[layer][i].coef * (currentOut * (1-currentOut) * esum));
					this.nodes[layer][i].calcABias(n);
					helper.writeDebug("going to calc aweights for "+layer+":"+i);
					this.nodes[layer][i].calcAWeights(n);
					helper.writeDebug("<<<<< "+i);
				}
			}

			//actually save the adjustments
			for(layer=0; layer<this.nodes.length; layer++){
				for(i=0; i<this.nodes[layer].length; i++){
					helper.writeDebug("saving changes for "+layer+":"+i);
					helper.writeDebug(JSON.stringify(this.nodes[layer][i].weights));
					helper.writeDebug(JSON.stringify(this.nodes[layer][i].aWeights));
					this.nodes[layer][i].applyAdjust();
				}
			}
		}
	}



	var nnNode = function(){  
		this.bias = 0;
		this.aBias = 0;
		this.output=null;
		this.errorVal=null;
		this.coef = 0.1;
		this.inputs=null;
		this.weights=null;
		this.isInputLayer=false;
		this.aWeights=Array();



		//expects an array
		this.setInputs = function(inputs){
			this.inputs = inputs;
		}

		this.setBias = function(bias){
			this.bias=bias;
		}

		//expects an array
		this.setWeights = function(weights){
			this.weights = weights;
		}

		this.setError = function(e){
			helper.writeDebug("error is "+e);
			this.errorVal=e;
		}

		this.checkForErrors = function(){
			if(this.inputs.length!=this.weights.length){
				return({"status":false, "msg":"weights!=inputs"});
			}
			return({"status":true})
		}

		this.calcOutput = function(){
			this.output=0;
			var key=0;
			for(key=0;key<this.inputs.length;key++){
				helper.writeDebug("about to do "+key+" element "+this.inputs[key]+" * "+this.weights[key]);
				this.output+=this.inputs[key] * this.weights[key];
			}
			this.output+=this.bias;
			this.output = (1/(1+Math.exp(-this.coef * this.output)));
			helper.writeDebug("did output = "+ this.output);

		}

		this.getOutput = function(){
			if(this.isInputLayer==false){
				this.calcOutput();
			}
			return (this.output);
		}


		this.applyAdjust = function(){
			//helper.writeDebug("saved new weights");
			this.bias = this.aBias;

			this.weights = this.aWeights;
			this.errorVal=-99999;
		}

		this.calcAWeights = function(n){
			var i;
			helper.writeDebug("attempting to get "+JSON.stringify(this.inputs));
			for(i=0; i<this.inputs.length; i++){
				this.aWeights[i] = this.weights[i] + (n * this.errorVal * this.inputs[i]);
				helper.writeDebug("calculating aweight: ("+this.weights[i]+" + ("+n+" * "+this.errorVal+" * "+ this.inputs[i]+") = " + this.aWeights[i]);

			}
		}

		this.calcABias = function(n){
			this.aBias = this.bias + (n * this.errorVal * 1);
		}

		this.getSingleError = function(wIndex){
			return(this.errorVal * this.weights[wIndex]);
		}
	};

	module.exports = nnNetwork;
